/*
 * Copyright (C) 2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blendb.
 *
 * blendb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blendb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blendb.  If not, see <http://www.gnu.org/licenses/>.
 */

import { expect } from "chai";
import { PostgresAdapter } from "./postgres";
import { MonetAdapter, MonetConfig } from "./monet";
import { Adapter } from "../core/adapter";
import { Fixture as FixPostgres } from "../../test/postgres/fixture";
import { Fixture as FixMonet } from "../../test/monet/fixture";
import { adapterScenario } from "../../test/scenario";
import { eachOf } from "async";

let adapters: Adapter[] = [];
let fixture;

function loadDb(db: string, index: number, cb: (err: any, result: Adapter) => void): void {
    let adapter: Adapter;

    if (db === "postgres") {
        fixture = new FixPostgres(adapterScenario.config.connections[index]);
        fixture.load(adapterScenario.config.loadViews, (err) => {
            if (err) {
                throw err;
            }
            adapter = new PostgresAdapter(adapterScenario.config.connections[index]);
            cb(null, adapter);
        });
    }
    else if (db === "monet") {
        fixture = new FixMonet(adapterScenario.config.connections[index]);
        fixture.load(adapterScenario.config.loadViews, (err) => {
            if (err) {
                throw err;
            }
            let parsedConfig: MonetConfig = {
                user: adapterScenario.config.connections[index].user,
                dbname: adapterScenario.config.connections[index].database,
                password: adapterScenario.config.connections[index].password,
                host: adapterScenario.config.connections[index].host,
                port: adapterScenario.config.connections[index].port
            };
            adapter = new MonetAdapter(parsedConfig);
            cb(null, adapter);
        });
    }
    else{
        cb("invalid adapter", null);
    }
}

describe("Sql adapter", () => {
    // Initializing
    before(function (done): void {
        // Arrow function not used to get acces to this and skip the test
        eachOf(adapterScenario.config.adapters , function(database, key: number, callback) {
            loadDb(database, key, function(err, result ) {
                if (err){
                    return callback(err);
                }
                else{
                    adapters[key] = result;
                }
                callback();
            });
        }, function (err){
            if (err){
                this.skip();
            }
            else{
                done();
            }
        });
    });

    for (let i = 0; i < adapterScenario.config.ndb; i++ ){

        it(adapterScenario.config.adapters[i] + ": " + "should get data from single materialized view", (done) => {
            let view = adapterScenario.materializedView;
            adapters[i].getDataFromView(view, (err, result) => {
                expect(err).to.be.a("null");
                expect(result).to.be.an("array");
                expect(result).to.have.length(5);
                let keys: string[] = [];
                keys = keys.concat(view.metrics.map((item) => item.name));
                keys = keys.concat(view.dimensions.map((item) => item.name));
                result.forEach((row) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
                done();
            });
        });
        it(adapterScenario.config.adapters[i] + ": " + "should sort data from join of two views", (done) => {
            let view = adapterScenario.sortView;
            adapters[i].getDataFromView(view, (err, result) => {
                expect(err).to.be.a("null");
                expect(result).to.be.an("array");
                expect(result).to.have.length(4);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(view.metrics.map((item) => item.name));
                keys = keys.concat(view.dimensions.map((item) => item.name));
                result.forEach((row) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
                done();
            });
        });
        it(adapterScenario.config.adapters[i] + ": " + "should get data from single view (with sub-dimension)", (done) => {
            let view = adapterScenario.subDimensionView;
            adapters[i].getDataFromView(view, (err, result) => {
                expect(err).to.be.a("null");
                expect(result).to.be.an("array");
                expect(result).to.have.length(5);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(view.metrics.map((item) => item.name));
                keys = keys.concat(view.dimensions.map((item) => item.name));
                result.forEach((row) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
                done();
            });
        });
        it(adapterScenario.config.adapters[i] + ": " + "should get data from join of one view", (done) => {
            let view = adapterScenario.joinWithOneView;
            adapters[i].getDataFromView(view, (err, result) => {
                expect(err).to.be.a("null");
                expect(result).to.be.an("array");
                expect(result).to.have.length(5);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(view.metrics.map((item) => item.name));
                keys = keys.concat(view.dimensions.map((item) => item.name));
                result.forEach((row) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
                done();
            });
        });
        it(adapterScenario.config.adapters[i] + ": " + "should get data from view with a single clause", (done) => {
        let view = adapterScenario.filterWithEqual;
        adapters[i].getDataFromView(view, (err, result) => {
                expect(err).to.be.a("null");
                expect(result).to.be.an("array");
                expect(result).to.have.length(1);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(view.metrics.map((item) => item.name));
                keys = keys.concat(view.dimensions.map((item) => item.name));
                result.forEach((row) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
                done();
            });
        });
        it(adapterScenario.config.adapters[i] + ": " + "should get data from reduce with the same attributes of view", (done) => {
            let view = adapterScenario.reduceAsView;
            adapters[i].getDataFromView(view, (err, result) => {
                expect(err).to.be.a("null");
                expect(result).to.be.an("array");
                expect(result).to.have.length(5);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(view.metrics.map((item) => item.name));
                keys = keys.concat(view.dimensions.map((item) => item.name));
                result.forEach((row) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
                done();
            });
        });
        it(adapterScenario.config.adapters[i] + ": " + "should get data from join into reduce with a single clause", (done) => {
            let view = adapterScenario.filterAverageBought;
            adapters[i].getDataFromView(view, (err, result) => {
                expect(err).to.be.a("null");
                expect(result).to.be.an("array");
                expect(result).to.have.length(3);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(view.metrics.map((item) => item.name));
                keys = keys.concat(view.dimensions.map((item) => item.name));
                result.forEach((row) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
                done();
            });
        });
        it(adapterScenario.config.adapters[i] + ": " + "should get data from view with multiple clauses", (done) => {
            let view = adapterScenario.multipleClause;
            adapters[i].getDataFromView(view, (err, result) => {
                expect(err).to.be.a("null");
                expect(result).to.be.an("array");
                expect(result).to.have.length(1);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(view.metrics.map((item) => item.name));
                keys = keys.concat(view.dimensions.map((item) => item.name));
                result.forEach((row) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
                done();
            });
        });
        it(adapterScenario.config.adapters[i] + ": " + "should get data from view with single clause, with more than one filter", (done) => {
            let view = adapterScenario.singleClause;
            adapters[i].getDataFromView(view, (err, result) => {
                expect(err).to.be.a("null");
                expect(result).to.be.an("array");
                expect(result).to.have.length(4);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(view.metrics.map((item) => item.name));
                keys = keys.concat(view.dimensions.map((item) => item.name));
                result.forEach((row) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
                done();
            });
        });
        it(adapterScenario.config.adapters[i] + ": " + "should get data from view with no metrics", (done) => {
            let view = adapterScenario.joinWithNoMetrics;
            adapters[i].getDataFromView(view, (err, result) => {
                expect(err).to.be.a("null");
                expect(result).to.be.an("array");
                expect(result).to.have.length(5);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(view.metrics.map((item) => item.name));
                keys = keys.concat(view.dimensions.map((item) => item.name));
                result.forEach((row) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
                done();
            });
        });
        it(adapterScenario.config.adapters[i] + ": " + "should get data from join with one metric and one view", (done) => {
            let view = adapterScenario.growOneView;
            adapters[i].getDataFromView(view, (err, result) => {
                expect(err).to.be.a("null");
                expect(result).to.be.an("array");
                expect(result).to.have.length(5);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(view.metrics.map((item) => item.name));
                keys = keys.concat(view.dimensions.map((item) => item.name));
                result.forEach((row) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
                done();
            });
        });
        it(adapterScenario.config.adapters[i] + ": " + "should get data from join with unrelated dimension", (done) => {
            let view = adapterScenario.JoinWithAncestors;
            adapters[i].getDataFromView(view, (err, result) => {
                expect(err).to.be.a("null");
                expect(result).to.be.an("array");
                expect(result).to.have.length(5);
                expect(result[0]).to.be.an("object");
                let keys: string[] = [];
                keys = keys.concat(view.metrics.map((item) => item.name));
                keys = keys.concat(view.dimensions.map((item) => item.name));
                result.forEach((row) => {
                    expect(row).to.be.an("object");
                    expect(row).to.have.all.keys(keys);
                });
                done();
            });
        });
    }
});
