/*
 * Copyright (C) 2016 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blend.
 *
 * blend is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blend is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blend.  If not, see <http://www.gnu.org/licenses/>.
 */

import { AggregationType, DataType } from "../common/types";
import { EnumHandler } from "../util/enumHandler";

/** Parameters used to create a metric object. */
export interface MetricOptions {
    /** Metric name. */
    name: string;
    /** Aggregation function applied on this metric */
    aggregation: AggregationType;
    /** Metric data type. */
    dataType: DataType;
    /** Breif description of what this metric represents. */
    description?: string;
}

/**
 * Parameters used to define metric object in the configuration file.
 * Also the string description of a metric.
 */
export interface MetricStrOptions {
    /** Metric name. */
    name: string;
    /** Aggregation function applied on this metric */
    aggregation: string;
    /** Metric data type. */
    dataType: string;
    /** Breif description of what this metric represents. */
    description?: string;
}

/**
 * One attribute of database that can be read.
 * A metric defines a numeric agggregated aspect of the data.
 * Used in a query as a desired information and is separated from
 * the dimension because this two types of attributes behave differently
 * in the query. While metrics are aggregated, dimensions are grouped.
 */
export class Metric {
    /** Metric name. */
    public readonly name: string;
    /** Aggregation function applied on this metric */
    public readonly aggregation: AggregationType;
    /** Metric data type. */
    public readonly dataType: DataType;
    /** Breif description of what this metric represents. */
    public readonly description: string;

    /**
     * Create a metric.
     * @param options - Parameters required to create a metric.
     */
    constructor(options: MetricOptions) {
        this.name = options.name;
        this.aggregation = options.aggregation;
        this.dataType = options.dataType;
        this.description = (options.description) ? options.description : "";
    }

    /**
     * Created a object with the same options used to create this
     * metric as strings. Used to inform the API users.
     */
    public strOptions(): MetricStrOptions {
        return {
            name: this.name,
            aggregation: EnumHandler.stringifyAggrType(this.aggregation),
            dataType: EnumHandler.stringfyDataType(this.dataType),
            description: this.description
        };
    }
}
