/*
 * Copyright (C) 2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blend.
 *
 * blend is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blend is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blend.  If not, see <http://www.gnu.org/licenses/>.
 */

import { Client, PoolConfig } from "pg";
import { View, LoadView } from "../../src/core/view";
import { Source } from "../../src/core/source";
import { each, series } from "async";
import * as fs from "fs";
import { DataType } from "../../src/common/types";

interface TransSet {
    init: string;
    data: string[];
}

interface SoucerParse{
    name: string;
    type: string[];
    fields: string[];
}

export interface Schema {
    alias?: string;
    query?: string;
    data?: string;
    fields: any[];
}

export class Fixture {
    private database: string;
    private config: PoolConfig;

    constructor(config: PoolConfig) {
        this.config = config;
        this.database = config.database;
    }

    public load(schemas: LoadView[], cb: (err: Error) => void): void {
        let client = new Client(this.config);

        /*
            Loading data has 2 steps:
            1 - Create a table or truncate a existing one.
            2 - Insert data.
        */
        let data: string[] = [];
        for (let i = 0; i < schemas.length; ++i) {
            data = data.concat(this.extractDataView(schemas[i].view, schemas[i].data));
        }
        client.connect((error) => {
            if (error) {
                cb(error);
                return;
            }
            /*
                Tables must be creates before data could be inserted, so the
                queries that create and insert are splited in 2 arrays.
                To garantee that tables exists before start insert data, series
                is used. Inside series each query is executed, using each;
            */
            series([(callback: (err: Error) => void) => {
                each(schemas, (query, cback) => {
                    client.query(this.createTransSet(query.view, query.data, true), [], (errCreate: Error) => {
                        if (errCreate) {
                            let regex = new RegExp(/relation \".*\" already exists/);
                            if (regex.test(errCreate.message.toLowerCase())){
                                return client.query(this.createTransSet(query.view, query.data, false), [], (errTrunc: Error) => cback(errTrunc));
                            }
                        }
                        return cback(errCreate);
                    });
                }, (errQuery: Error) => callback(errQuery));
            }, (callback: (err: Error) => void) => {
                each(data, (query, cback) => {
                    return client.query( query, [], (err: Error) => cback(err));
                }, (errQuery: Error) => callback(errQuery));
            }], (errQuery: Error) => {
                if (errQuery) {
                    client.end();
                    cb(errQuery);
                }
                client.end((err) => {
                    cb(err);
                });
            });

        });
    }

    private typeConvertion(t: DataType): string {
        switch (t) {
            case DataType.INTEGER:
                return "INTEGER";
            case DataType.DATE:
                return "DATE";
            case DataType.STRING:
                return "TEXT";
            case DataType.BOOLEAN:
                return "BOOLEAN";
            case DataType.FLOAT:
                return "DOUBLE PRECISION";
            default:
                return "TEXT";
        }
    }

    private extractDataView(view: View, filePath: string): string[]{
        let name = view.name;
        let data: string[] = [];
        let rows = JSON.parse(fs.readFileSync(filePath, {encoding : "utf8"}));
        for (let i = 0; i < rows.length; ++i) {
            let values = [];
            let keys = [];
            for (let key of Object.keys(rows[i])) {
                keys.push("\"" + key + "\"");
                values.push("'" + rows[i][key] + "'");
            }
            data.push("INSERT INTO " + name +
                                    "(" + keys.join(", ") + ") " +
                                    "VALUES (" + values.join(", ") + ")");
        }

        return data;

    }

    private createTransSet(view: View, filePath: string, create: boolean): string {
        let name = view.name;
        let init: string = "";
        let props = [];
        for (let i = 0; i < view.metrics.length; ++i) {
            let met = view.metrics[i];
            props.push("\"" + met.name + "\" " + this.typeConvertion(met.dataType));
        }
        for (let i = 0; i < view.dimensions.length; ++i) {
            let dim = view.dimensions[i];
            props.push("\"" + dim.name + "\" " + this.typeConvertion(dim.dataType));
        }
        if (create) {
            init = "CREATE TABLE " + name + "(" + props.join(", ") + ")";
        }
        else {
            init = "TRUNCATE TABLE " + name;
        }
        return init;
    }
    public loadSource(source: Source[] , cb: (err: Error) => void): void {

        let client = new Client(this.config);

        client.connect((error) => {
            if (error) {
                cb(error);
                return;
            }
            series([(callback: (err: Error) => void) => {
                each(source, (insere, cback) => {
                    client.query ( this.extractData(insere, true), [], (errCreate: Error) => {
                        if (errCreate) {
                            let regex = new RegExp(/relation \".*\" already exists/);
                            if (regex.test(errCreate.message.toLowerCase())){
                                return client.query ( this.extractData(insere, false), [], (errTrunc: Error) => cback(errTrunc));
                            }
                        }
                        return cback(errCreate);
                    });
                }, (errQuery: Error) => callback(errQuery));
            }], (errQuery: Error) => {
                 if (errQuery) {
                     client.end();
                     cb(errQuery);
                 }
                 client.end((err) => {
                     cb(err);
                 });
            });

        });
    }
    private extractData(data: Source, create: boolean): string{
        let name: string;
        let type: DataType[];
        let fields: string[];
        let consult: string;

        name = data.name;
        type = data.fields.map((item) => item.dataType);
        fields = data.fields.map((item) => item.name);

        if (create){
        consult = "CREATE TABLE " + name + " (" + '"';
        for (let i = 0; i < fields.length; i++){
            fields[i] = fields[i].concat('"' + " " + this.typeConvertion(type[i]));
        }
        consult = consult.concat(fields.join(", " + '"'));
        consult = consult.concat(");");
        }
        else{
        consult = "TRUNCATE TABLE " + name + ";";
        }

        return consult;

    }
}
